Zodish.Type.Date (zodish v0.2.4)

This module describes a Zodish date type.

Summary

Functions

Either enables or disables coercion for the given Date type.

Creates a new Date type.

Types

t()

@type t() :: %Zodish.Type.Date{coerce: boolean()}

Functions

coerce(type, value \\ true)

Either enables or disables coercion for the given Date type.

new(opts \\ [])

Creates a new Date type.