Zodish.Type.Transform (zodish v0.2.4)

Summary

Functions

Applies a one-arity transformation function or MFA to the output of the given inner type.

Types

t()

@type t() :: t(Zodish.Type.t())

t(inner_type)

@type t(inner_type) :: %Zodish.Type.Transform{
  fun: (any() -> any()) | mfa(),
  inner_type: inner_type
}

Functions

new(inner_type, fun)

Applies a one-arity transformation function or MFA to the output of the given inner type.