Zodish.Type protocol (zodish v0.2.4)
View SourceProtocol for parsing values based on Zodish types.
Summary
Functions
Parses a value based on the given Zodish type.
Builds the type spec for a given Zodish type.
Types
@type t() :: term()
All the types that implement this protocol.