Zodish.Type.Enum (zodish v0.2.4)
View SourceThis module describes a Zodish enum type (atoms only).
Summary
Functions
Either enables or disables coercion for the given Enum (atom) type.
Creates a new Enum type.
This module describes a Zodish enum type (atoms only).
Either enables or disables coercion for the given Enum (atom) type.
Creates a new Enum type.