Zodish.Type.Transform (zodish v0.2.4)
View SourceSummary
Functions
Applies a one-arity transformation function or MFA to the output of the given inner type.
Types
@type t() :: t(Zodish.Type.t())
Applies a one-arity transformation function or MFA to the output of the given inner type.
@type t() :: t(Zodish.Type.t())